from __future__ import annotations  # Postponed Evaluation of Annotations
from typing import TYPE_CHECKING, Callable, Optional
from uuid import UUID, uuid4
from cumo._internal.protobuf import server_pb2
if TYPE_CHECKING:
    from cumo import PointCloudViewer


# pylint: disable=no-member, redefined-builtin


def add_custom_slider(
    self: PointCloudViewer,
    name: str = "slider",
    min_value: float = 0,
    max_value: float = 100,
    step: float = 1,
    init_value: float = 50,
    on_changed: Optional[Callable[[float], None]] = None,
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダにスライダーを追加する。

    :param name: 表示名
    :type name: str, optional
    :param min_value: 最小値
    :type min_value: float, optional
    :param max_value: 最大値
    :type max_value: float, optional
    :param step: 最小の変化量
    :type step: float, optional
    :param init_value: 初期値
    :type init_value: float, optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にスライダーの設定値が渡される
    :type on_changed: Optional[Callable[[float], None]], optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    obj = server_pb2.ServerCommand()
    slider = server_pb2.CustomControl.Slider()
    slider.name = name
    slider.min = min_value
    slider.max = max_value
    slider.step = step
    slider.init_value = init_value
    slider.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.slider.CopyFrom(slider)
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._set_custom_handler(uuid, "changed", on_changed)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def add_custom_checkbox(
    self: PointCloudViewer,
    name: str = "checkbox",
    init_value: bool = False,
    on_changed: Optional[Callable[[bool], None]] = None,
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダーにチェックボックスを追加する。

    :param name: 表示名
    :type name: str, optional
    :param init_value: 初期値
    :type init_value: bool, optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にチェックボックスの設定値が渡される
    :type on_changed: Optional[Callable[[bool], None]], optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional


    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    checkbox = server_pb2.CustomControl.CheckBox()
    checkbox.name = name
    checkbox.init_value = init_value
    checkbox.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.checkbox.CopyFrom(checkbox)
    obj = server_pb2.ServerCommand()
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._set_custom_handler(uuid, "changed", on_changed)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def add_custom_textbox(
    self: PointCloudViewer,
    name: str = "textbox",
    init_value: str = "",
    on_changed: Optional[Callable[[bool], None]] = None,
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダーにテキストボックスを追加する。

    :param name: 表示名
    :type name: str, optional
    :param init_value: 初期値
    :type init_value: str, optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にテキストボックスの値が渡される
    :type on_changed: Optional[Callable[[bool], None]], optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    textbox = server_pb2.CustomControl.TextBox()
    textbox.name = name
    textbox.init_value = init_value
    textbox.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.textbox.CopyFrom(textbox)
    obj = server_pb2.ServerCommand()
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._set_custom_handler(uuid, "changed", on_changed)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def add_custom_selectbox(
    self: PointCloudViewer,
    items: list,
    name: str = "selectbox",
    init_value: str = "",
    on_changed: Optional[Callable[[bool], None]] = None,
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダーにセレクトボックスを追加する。

    :param items: 要素の文字列のリスト
    :type items: list
    :param name: 表示名
    :type name: str, optional
    :param init_value: 初期値
    :type init_value: str, optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数に選択された要素の文字列が渡される
    :type on_changed: Optional[Callable[[bool], None]], optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    selectbox = server_pb2.CustomControl.SelectBox()
    selectbox.name = name
    selectbox.items.extend(items)
    selectbox.init_value = init_value
    selectbox.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.selectbox.CopyFrom(selectbox)
    obj = server_pb2.ServerCommand()
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._set_custom_handler(uuid, "changed", on_changed)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def add_custom_button(
    self: PointCloudViewer,
    name: str = "button",
    on_changed: Optional[Callable[[bool], None]] = None,
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダーにボタンを追加する。

    :param name: 表示名
    :type name: str, optional
    :param on_changed: ボタンが押されたときに呼ばれるコールバック関数。引数に ``True`` が渡される
    :type on_changed: Optional[Callable[[bool], None]], optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    button = server_pb2.CustomControl.Button()
    button.name = name
    button.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.button.CopyFrom(button)
    obj = server_pb2.ServerCommand()
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._set_custom_handler(uuid, "changed", on_changed)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def add_custom_colorpicker(
    self: PointCloudViewer,
    name: str = "color",
    init_value: str = "#000",
    on_changed: Optional[Callable[[], None]] = None,
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダーにカラーピッカーを追加する。

    :param name: 表示名
    :type name: str, optional
    :param init_value: 初期値。 ``#ff0000`` 、 ``rgb(255,0,0)`` 等の表現が利用可能
    :type init_value: str, optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数に色を表す文字列が渡される
    :type on_changed: Optional[Callable[[], None]], optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    picker = server_pb2.CustomControl.ColorPicker()
    picker.name = name
    picker.init_value = init_value
    picker.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.color_picker.CopyFrom(picker)
    obj = server_pb2.ServerCommand()
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._set_custom_handler(uuid, "changed", on_changed)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def remove_all_custom_controls(
    self: PointCloudViewer
) -> None:
    """すべてのカスタムコントロールを削除する。
    """
    remove_custom_control_cmd = server_pb2.RemoveCustomControl()
    remove_custom_control_cmd.all = True

    obj = server_pb2.ServerCommand()
    obj.remove_custom_control.CopyFrom(remove_custom_control_cmd)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def remove_custom_control(
    self: PointCloudViewer,
    uuid: UUID
) -> None:
    """指定したUUIDを持つカスタムコントロールを削除する。

    Args:
        uuid (UUID): 削除するコントロールに対応するID
    """
    remove_custom_control_cmd = server_pb2.RemoveCustomControl()
    remove_custom_control_cmd.by_uuid = str(uuid)

    obj = server_pb2.ServerCommand()
    obj.remove_custom_control.CopyFrom(remove_custom_control_cmd)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def add_custom_folder(
    self: PointCloudViewer,
    name: str = "folder",
    parent: UUID = UUID(int=0),
) -> UUID:
    """カスタムフォルダーにフォルダーを追加する。

    :param name: 表示名
    :type name: str, optional
    :param parent: 親フォルダーのUUID。特に指定しない場合はルートフォルダーになる。
    :type parent: UUID, optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    folder = server_pb2.CustomControl.Folder()
    folder.name = name
    folder.parent = str(parent)
    add_custom_control = server_pb2.CustomControl()
    add_custom_control.folder.CopyFrom(folder)
    obj = server_pb2.ServerCommand()
    obj.add_custom_control.CopyFrom(add_custom_control)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)



__all__ = ["PointCloudViewer"]

import multiprocessing
from enum import Enum, auto
from typing import Optional

# pylint: disable=import-outside-toplevel


class DownSampleStrategy(Enum):
    NONE = auto()
    RANDOM_SAMPLE = auto()
    VOXEL_GRID = auto()

    voxel_size: Optional[float]

    def set_voxel_size(self, voxel_size):
        self.voxel_size = voxel_size
        return self


class PointCloudViewer:
    """点群をブラウザで表示するためのサーバーを立ち上げるビューア。

    :param host: ホスト名
    :type host: str, optional
    :param websocket_port: WebSocketサーバーが待ち受けるポート
    :type websocket_port: int, optional
    :param http_port: Webサーバーが待ち受けるポート
    :type http_port: int, optional
    :param polling_interval: サーバープロセスとのやり取りをする間隔の秒数
    :type polling_interval: int, optional
    :param autostart: Trueの場合、 ``start`` がコンストラクタ実行時に呼び出される
    :type autostart: bool, optional
    """
    _server_process: multiprocessing.Process
    _custom_handlers: dict
    _key_event_handlers: dict
    _websocket_broadcasting_queue: multiprocessing.Queue
    _websocket_message_queue: multiprocessing.Queue

    from cumo._internal.members.capture_screen import (
        capture_screen,
        capture_screen_as_ndarray,
    )
    from cumo._internal.members.camera import (
        set_camera_position,
        set_camera_target,
        set_orthographic_camera,
        set_perspective_camera,
    )
    from cumo._internal.members.custom_control import (
        add_custom_button,
        add_custom_checkbox,
        add_custom_colorpicker,
        add_custom_selectbox,
        add_custom_slider,
        add_custom_textbox,
        add_custom_folder,
        remove_all_custom_controls,
        remove_custom_control,
    )
    from cumo._internal.members.internal_utils import __init__
    from cumo._internal.members.send_object import (
        send_lineset,
        send_overlay_text,
        send_overlay_image,
        send_overlay_image_from_ndarray,
        send_pointcloud,
        send_pointcloud_pcd,
        send_mesh,
    )
    from cumo._internal.members.utils import (
        wait_forever,
        console_log,
        start,
    )
    from cumo._internal.members.event_handler import (
        _get_custom_handler,
        _handle_message,
        _send_data,
        _set_custom_handler,
        _wait_until,
    )
    from cumo._internal.members.keyboard_event_handler import (
        add_keyup_handler,
        remove_keyup_handler,
        add_keydown_handler,
        remove_keydown_handler,
        add_keypress_handler,
        remove_keypress_handler,
    )
    from cumo._internal.members.remove_object import (
        remove_all_objects,
        remove_object,
    )
    from cumo._internal.members.set_custom_control import (
        set_custom_slider,
        set_custom_selectbox,
        set_custom_colorpicker,
        set_custom_textbox,
        set_custom_checkbox,
        set_custom_button,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf/client.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf/client.proto',
  package='Client',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15protobuf/client.proto\x12\x06\x43lient\"\xec\x01\n\rClientCommand\x12\x0c\n\x04UUID\x18\x01 \x01(\t\x12\x15\n\x0blog_message\x18\x02 \x01(\tH\x00\x12 \n\x06result\x18\x03 \x01(\x0b\x32\x0e.Client.ResultH\x00\x12\x1e\n\x05image\x18\x04 \x01(\x0b\x32\r.Client.ImageH\x00\x12\x31\n\x0f\x63ontrol_changed\x18\x05 \x01(\x0b\x32\x16.Client.ControlChangedH\x00\x12\x36\n\x12key_event_occurred\x18\x06 \x01(\x0b\x32\x18.Client.KeyEventOccurredH\x00\x42\t\n\x07\x43ommand\"8\n\x06Result\x12\x11\n\x07success\x18\x01 \x01(\tH\x00\x12\x11\n\x07\x66\x61ilure\x18\x02 \x01(\tH\x00\x42\x08\n\x06Result\"\x15\n\x05Image\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"N\n\x0e\x43ontrolChanged\x12\x10\n\x06number\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04text\x18\x02 \x01(\tH\x00\x12\x11\n\x07\x62oolean\x18\x03 \x01(\x08H\x00\x42\x07\n\x05Value\"\xb7\x02\n\x10KeyEventOccurred\x12\x32\n\x05keyup\x18\x01 \x01(\x0b\x32!.Client.KeyEventOccurred.KeyEventH\x00\x12\x34\n\x07keydown\x18\x02 \x01(\x0b\x32!.Client.KeyEventOccurred.KeyEventH\x00\x12\x35\n\x08keypress\x18\x03 \x01(\x0b\x32!.Client.KeyEventOccurred.KeyEventH\x00\x1ay\n\x08KeyEvent\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x10\n\x08shiftKey\x18\x03 \x01(\x08\x12\x0e\n\x06\x61ltKey\x18\x04 \x01(\x08\x12\x0f\n\x07\x63trlKey\x18\x05 \x01(\x08\x12\x0f\n\x07metaKey\x18\x06 \x01(\x08\x12\x0e\n\x06repeat\x18\x07 \x01(\x08\x42\x07\n\x05\x45ventb\x06proto3'
)




_CLIENTCOMMAND = _descriptor.Descriptor(
  name='ClientCommand',
  full_name='Client.ClientCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='UUID', full_name='Client.ClientCommand.UUID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='log_message', full_name='Client.ClientCommand.log_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result', full_name='Client.ClientCommand.result', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image', full_name='Client.ClientCommand.image', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='control_changed', full_name='Client.ClientCommand.control_changed', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_event_occurred', full_name='Client.ClientCommand.key_event_occurred', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Command', full_name='Client.ClientCommand.Command',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=34,
  serialized_end=270,
)


_RESULT = _descriptor.Descriptor(
  name='Result',
  full_name='Client.Result',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='Client.Result.success', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='failure', full_name='Client.Result.failure', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Result', full_name='Client.Result.Result',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=272,
  serialized_end=328,
)


_IMAGE = _descriptor.Descriptor(
  name='Image',
  full_name='Client.Image',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='Client.Image.data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=330,
  serialized_end=351,
)


_CONTROLCHANGED = _descriptor.Descriptor(
  name='ControlChanged',
  full_name='Client.ControlChanged',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='number', full_name='Client.ControlChanged.number', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='Client.ControlChanged.text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boolean', full_name='Client.ControlChanged.boolean', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Value', full_name='Client.ControlChanged.Value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=353,
  serialized_end=431,
)


_KEYEVENTOCCURRED_KEYEVENT = _descriptor.Descriptor(
  name='KeyEvent',
  full_name='Client.KeyEventOccurred.KeyEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Client.KeyEventOccurred.KeyEvent.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='Client.KeyEventOccurred.KeyEvent.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shiftKey', full_name='Client.KeyEventOccurred.KeyEvent.shiftKey', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='altKey', full_name='Client.KeyEventOccurred.KeyEvent.altKey', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ctrlKey', full_name='Client.KeyEventOccurred.KeyEvent.ctrlKey', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metaKey', full_name='Client.KeyEventOccurred.KeyEvent.metaKey', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repeat', full_name='Client.KeyEventOccurred.KeyEvent.repeat', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=615,
  serialized_end=736,
)

_KEYEVENTOCCURRED = _descriptor.Descriptor(
  name='KeyEventOccurred',
  full_name='Client.KeyEventOccurred',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='keyup', full_name='Client.KeyEventOccurred.keyup', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keydown', full_name='Client.KeyEventOccurred.keydown', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keypress', full_name='Client.KeyEventOccurred.keypress', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_KEYEVENTOCCURRED_KEYEVENT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Event', full_name='Client.KeyEventOccurred.Event',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=434,
  serialized_end=745,
)

_CLIENTCOMMAND.fields_by_name['result'].message_type = _RESULT
_CLIENTCOMMAND.fields_by_name['image'].message_type = _IMAGE
_CLIENTCOMMAND.fields_by_name['control_changed'].message_type = _CONTROLCHANGED
_CLIENTCOMMAND.fields_by_name['key_event_occurred'].message_type = _KEYEVENTOCCURRED
_CLIENTCOMMAND.oneofs_by_name['Command'].fields.append(
  _CLIENTCOMMAND.fields_by_name['log_message'])
_CLIENTCOMMAND.fields_by_name['log_message'].containing_oneof = _CLIENTCOMMAND.oneofs_by_name['Command']
_CLIENTCOMMAND.oneofs_by_name['Command'].fields.append(
  _CLIENTCOMMAND.fields_by_name['result'])
_CLIENTCOMMAND.fields_by_name['result'].containing_oneof = _CLIENTCOMMAND.oneofs_by_name['Command']
_CLIENTCOMMAND.oneofs_by_name['Command'].fields.append(
  _CLIENTCOMMAND.fields_by_name['image'])
_CLIENTCOMMAND.fields_by_name['image'].containing_oneof = _CLIENTCOMMAND.oneofs_by_name['Command']
_CLIENTCOMMAND.oneofs_by_name['Command'].fields.append(
  _CLIENTCOMMAND.fields_by_name['control_changed'])
_CLIENTCOMMAND.fields_by_name['control_changed'].containing_oneof = _CLIENTCOMMAND.oneofs_by_name['Command']
_CLIENTCOMMAND.oneofs_by_name['Command'].fields.append(
  _CLIENTCOMMAND.fields_by_name['key_event_occurred'])
_CLIENTCOMMAND.fields_by_name['key_event_occurred'].containing_oneof = _CLIENTCOMMAND.oneofs_by_name['Command']
_RESULT.oneofs_by_name['Result'].fields.append(
  _RESULT.fields_by_name['success'])
_RESULT.fields_by_name['success'].containing_oneof = _RESULT.oneofs_by_name['Result']
_RESULT.oneofs_by_name['Result'].fields.append(
  _RESULT.fields_by_name['failure'])
_RESULT.fields_by_name['failure'].containing_oneof = _RESULT.oneofs_by_name['Result']
_CONTROLCHANGED.oneofs_by_name['Value'].fields.append(
  _CONTROLCHANGED.fields_by_name['number'])
_CONTROLCHANGED.fields_by_name['number'].containing_oneof = _CONTROLCHANGED.oneofs_by_name['Value']
_CONTROLCHANGED.oneofs_by_name['Value'].fields.append(
  _CONTROLCHANGED.fields_by_name['text'])
_CONTROLCHANGED.fields_by_name['text'].containing_oneof = _CONTROLCHANGED.oneofs_by_name['Value']
_CONTROLCHANGED.oneofs_by_name['Value'].fields.append(
  _CONTROLCHANGED.fields_by_name['boolean'])
_CONTROLCHANGED.fields_by_name['boolean'].containing_oneof = _CONTROLCHANGED.oneofs_by_name['Value']
_KEYEVENTOCCURRED_KEYEVENT.containing_type = _KEYEVENTOCCURRED
_KEYEVENTOCCURRED.fields_by_name['keyup'].message_type = _KEYEVENTOCCURRED_KEYEVENT
_KEYEVENTOCCURRED.fields_by_name['keydown'].message_type = _KEYEVENTOCCURRED_KEYEVENT
_KEYEVENTOCCURRED.fields_by_name['keypress'].message_type = _KEYEVENTOCCURRED_KEYEVENT
_KEYEVENTOCCURRED.oneofs_by_name['Event'].fields.append(
  _KEYEVENTOCCURRED.fields_by_name['keyup'])
_KEYEVENTOCCURRED.fields_by_name['keyup'].containing_oneof = _KEYEVENTOCCURRED.oneofs_by_name['Event']
_KEYEVENTOCCURRED.oneofs_by_name['Event'].fields.append(
  _KEYEVENTOCCURRED.fields_by_name['keydown'])
_KEYEVENTOCCURRED.fields_by_name['keydown'].containing_oneof = _KEYEVENTOCCURRED.oneofs_by_name['Event']
_KEYEVENTOCCURRED.oneofs_by_name['Event'].fields.append(
  _KEYEVENTOCCURRED.fields_by_name['keypress'])
_KEYEVENTOCCURRED.fields_by_name['keypress'].containing_oneof = _KEYEVENTOCCURRED.oneofs_by_name['Event']
DESCRIPTOR.message_types_by_name['ClientCommand'] = _CLIENTCOMMAND
DESCRIPTOR.message_types_by_name['Result'] = _RESULT
DESCRIPTOR.message_types_by_name['Image'] = _IMAGE
DESCRIPTOR.message_types_by_name['ControlChanged'] = _CONTROLCHANGED
DESCRIPTOR.message_types_by_name['KeyEventOccurred'] = _KEYEVENTOCCURRED
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ClientCommand = _reflection.GeneratedProtocolMessageType('ClientCommand', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTCOMMAND,
  '__module__' : 'protobuf.client_pb2'
  # @@protoc_insertion_point(class_scope:Client.ClientCommand)
  })
_sym_db.RegisterMessage(ClientCommand)

Result = _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
  'DESCRIPTOR' : _RESULT,
  '__module__' : 'protobuf.client_pb2'
  # @@protoc_insertion_point(class_scope:Client.Result)
  })
_sym_db.RegisterMessage(Result)

Image = _reflection.GeneratedProtocolMessageType('Image', (_message.Message,), {
  'DESCRIPTOR' : _IMAGE,
  '__module__' : 'protobuf.client_pb2'
  # @@protoc_insertion_point(class_scope:Client.Image)
  })
_sym_db.RegisterMessage(Image)

ControlChanged = _reflection.GeneratedProtocolMessageType('ControlChanged', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLCHANGED,
  '__module__' : 'protobuf.client_pb2'
  # @@protoc_insertion_point(class_scope:Client.ControlChanged)
  })
_sym_db.RegisterMessage(ControlChanged)

KeyEventOccurred = _reflection.GeneratedProtocolMessageType('KeyEventOccurred', (_message.Message,), {

  'KeyEvent' : _reflection.GeneratedProtocolMessageType('KeyEvent', (_message.Message,), {
    'DESCRIPTOR' : _KEYEVENTOCCURRED_KEYEVENT,
    '__module__' : 'protobuf.client_pb2'
    # @@protoc_insertion_point(class_scope:Client.KeyEventOccurred.KeyEvent)
    })
  ,
  'DESCRIPTOR' : _KEYEVENTOCCURRED,
  '__module__' : 'protobuf.client_pb2'
  # @@protoc_insertion_point(class_scope:Client.KeyEventOccurred)
  })
_sym_db.RegisterMessage(KeyEventOccurred)
_sym_db.RegisterMessage(KeyEventOccurred.KeyEvent)


# @@protoc_insertion_point(module_scope)



from __future__ import annotations  # Postponed Evaluation of Annotations
import queue
import base64
from typing import TYPE_CHECKING, Callable, Optional
from uuid import UUID
from google.protobuf.message import DecodeError
from cumo._internal.protobuf import server_pb2
from cumo._internal.protobuf import client_pb2
from cumo.keyboard_event import KeyboardEvent
if TYPE_CHECKING:
    from cumo import PointCloudViewer

# pylint: disable=no-member


def clean_defer_queue(self: PointCloudViewer, q: queue.Queue):
    while not q.empty():
        self._websocket_message_queue.put(q.get())


def _wait_until(self: PointCloudViewer, uuid: UUID) -> client_pb2.ClientCommand:
    defer_queue = queue.Queue()
    while True:
        data: bytes = self._websocket_message_queue.get()
        command: client_pb2.ClientCommand = client_pb2.ClientCommand()
        try:
            command.ParseFromString(data)
        except DecodeError as decode_error:
            clean_defer_queue(self, defer_queue)
            raise RuntimeError("failed to parsing message") from decode_error
        if UUID(hex=command.UUID) == uuid:
            clean_defer_queue(self, defer_queue)
            return command
        for_me = self._handle_message(command)
        if not for_me:
            # 他の_wait_untilループで受け取るべきデータが来てしまったので退避させる
            defer_queue.put(data)


def _handle_message(self: PointCloudViewer, command: client_pb2.ClientCommand) -> bool:
    if command.HasField("control_changed"):
        _handle_control_changed(self, command)
        return True
    if command.HasField("key_event_occurred"):
        _handle_key_event_occurred(self, command)
        return True
    return False


def _get_custom_handler(self: PointCloudViewer, uuid: UUID, name: str) -> Optional[Callable]:
    t = (uuid, name)
    if t in self._custom_handlers:
        return self._custom_handlers[t]
    return None


def _set_custom_handler(self: PointCloudViewer, uuid: UUID, name: str, func: Callable) -> None:
    self._custom_handlers[(uuid, name)] = func


def _handle_control_changed(self: PointCloudViewer, command: client_pb2.ClientCommand):
    uuid = UUID(hex=command.UUID)
    on_changed = self._get_custom_handler(uuid, "changed")
    if on_changed is not None:
        if command.control_changed.HasField("number"):
            on_changed(command.control_changed.number)
        elif command.control_changed.HasField("text"):
            on_changed(command.control_changed.text)
        elif command.control_changed.HasField("boolean"):
            on_changed(command.control_changed.boolean)


def _handle_key_event_occurred(self: PointCloudViewer, command: client_pb2.ClientCommand):
    if command.key_event_occurred.HasField("keyup"):
        handle_keyup(self, command.key_event_occurred.keyup)
    elif command.key_event_occurred.HasField("keydown"):
        handle_keydown(self, command.key_event_occurred.keydown)
    elif command.key_event_occurred.HasField("keypress"):
        handle_keypress(self, command.key_event_occurred.keypress)


def protobuf_to_keyboardevent(ev_pb: client_pb2.KeyEventOccurred.KeyEvent) -> KeyboardEvent:
    return KeyboardEvent(
        key=ev_pb.key,
        code=ev_pb.code,
        shiftKey=ev_pb.shiftKey,
        altKey=ev_pb.altKey,
        ctrlKey=ev_pb.ctrlKey,
        metaKey=ev_pb.metaKey,
        repeat=ev_pb.repeat,
    )


def handle_keyup(self: PointCloudViewer, ev_pb: client_pb2.KeyEventOccurred.KeyEvent):
    ev = protobuf_to_keyboardevent(ev_pb)
    items = list(self._key_event_handlers["keyup"].items())
    for t in items:
        uuid: UUID = t[0]
        handler: Callable[[KeyboardEvent, UUID], None] = t[1]
        handler(ev, uuid)


def handle_keydown(self: PointCloudViewer, ev_pb: client_pb2.KeyEventOccurred.KeyEvent):
    ev = protobuf_to_keyboardevent(ev_pb)
    items = list(self._key_event_handlers["keydown"].items())
    for t in items:
        uuid: UUID = t[0]
        handler: Callable[[KeyboardEvent, UUID], None] = t[1]
        handler(ev, uuid)


def handle_keypress(self: PointCloudViewer, ev_pb: client_pb2.KeyEventOccurred.KeyEvent):
    ev = protobuf_to_keyboardevent(ev_pb)
    items = list(self._key_event_handlers["keypress"].items())
    for t in items:
        uuid: UUID = t[0]
        handler: Callable[[KeyboardEvent, UUID], None] = t[1]
        handler(ev, uuid)


def _send_data(self: PointCloudViewer, pbobj: server_pb2.ServerCommand, uuid: UUID) -> None:
    pbobj.UUID = str(uuid)
    data = base64.b64encode(pbobj.SerializeToString())
    self._websocket_broadcasting_queue.put(data.decode())



from __future__ import annotations  # Postponed Evaluation of Annotations
from io import BytesIO
from typing import TYPE_CHECKING
from uuid import uuid4
from numpy import ndarray
import numpy as np
from PIL import Image
from cumo._internal.protobuf import server_pb2
if TYPE_CHECKING:
    from cumo import PointCloudViewer


# pylint: disable=no-member


def capture_screen(
    self: PointCloudViewer,
) -> bytes:
    """ブラウザのcanvasに表示されている画像をpng形式で保存させる。
    """
    obj = server_pb2.ServerCommand()
    obj.capture_screen = True
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if ret.HasField("image"):
        return ret.image.data
    raise RuntimeError("Unreachable")


def capture_screen_as_ndarray(
    self: PointCloudViewer,
) -> ndarray:
    """ブラウザのcanvasに表示されている画像を、shape が (height,width,3) で dtype が uint8 の ndarray で返す。
    """
    data = self.capture_screen()
    img = Image.open(BytesIO(data)).convert("RGB")
    return np.array(img)



from __future__ import annotations  # Postponed Evaluation of Annotations
from uuid import uuid4
from typing import TYPE_CHECKING
from cumo._internal.protobuf import server_pb2
if TYPE_CHECKING:
    from cumo import PointCloudViewer

# pylint: disable=no-member


def set_orthographic_camera(
    self: PointCloudViewer,
    frustum_height: float = 30,
) -> None:
    """カメラを正投影カメラに切り替えさせる。

    :param frustum_height: カメラの視錐台の高さ。幅はウィンドウのアスペクト比から計算される
    :type frustum_height: float, optional
    """
    camera = server_pb2.SetCamera()
    camera.orthographic_frustum_height = frustum_height
    obj = server_pb2.ServerCommand()
    obj.set_camera.CopyFrom(camera)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def set_perspective_camera(
    self: PointCloudViewer,
    fov: float = 30,
) -> None:
    """カメラを遠近投影カメラに切り替えさせる。

    :param fov: 視野角
    :type fov: float, optional
    """
    camera = server_pb2.SetCamera()
    camera.perspective_fov = fov
    obj = server_pb2.ServerCommand()
    obj.set_camera.CopyFrom(camera)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def set_camera_position(
    self: PointCloudViewer,
    x: float,
    y: float,
    z: float,
) -> None:
    """カメラの位置を変更する。

    :param x: カメラのx座標
    :type x: float
    :param y: カメラのy座標
    :type y: float
    :param z: カメラのz座標
    :type z: float
    """
    position = server_pb2.VecXYZf()
    position.x = x
    position.y = y
    position.z = z

    camera = server_pb2.SetCamera()
    camera.position.CopyFrom(position)

    obj = server_pb2.ServerCommand()
    obj.set_camera.CopyFrom(camera)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def set_camera_target(
    self: PointCloudViewer,
    x: float,
    y: float,
    z: float,
) -> None:
    """カメラが向く目標の座標を変更する。

    :param x: 目標のx座標
    :type x: float
    :param y: 目標のy座標
    :type y: float
    :param z: 目標のz座標
    :type z: float
    """
    target = server_pb2.VecXYZf()
    target.x = x
    target.y = y
    target.z = z

    camera = server_pb2.SetCamera()
    camera.target.CopyFrom(target)

    obj = server_pb2.ServerCommand()
    obj.set_camera.CopyFrom(camera)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)



from __future__ import annotations  # Postponed Evaluation of Annotations
from uuid import UUID, uuid4
from typing import Callable, TYPE_CHECKING
from cumo._internal.protobuf import server_pb2
from cumo.keyboard_event import KeyboardEvent
if TYPE_CHECKING:
    from cumo import PointCloudViewer

# pylint: disable=no-member


def add_keyup_handler(
    self: PointCloudViewer,
    handler: Callable[[KeyboardEvent, UUID], None]
) -> UUID:
    """ブラウザでkeyupイベントが発生したときに呼ばれるハンドラーを登録する。

    Args:
        handler (Callable[[KeyboardEvent, UUID], None]): ブラウザでkeyupイベントが発生したときに呼ばれるハンドラー

    Returns:
        UUID: ハンドラーに対応するID。後から操作する際に使う
    """
    uuid = uuid4()
    if "keyup" not in self._key_event_handlers:
        self._key_event_handlers["keyup"] = {}
    self._key_event_handlers["keyup"][uuid] = handler

    set_key_event_handler = server_pb2.SetKeyEventHandler()
    set_key_event_handler.keyup = True
    obj = server_pb2.ServerCommand()
    obj.set_key_event_handler.CopyFrom(set_key_event_handler)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    return uuid


def remove_keyup_handler(
    self: PointCloudViewer,
    uuid: UUID
) -> None:
    """ブラウザでkeyupイベントが発生したときに呼ばれるハンドラーを削除する。

    Args:
        uuid (UUID): ブラウザでkeyupイベントが発生したときに呼ばれるハンドラーのID
    """
    if "keyup" not in self._key_event_handlers:
        raise KeyError(uuid)
    if uuid not in self._key_event_handlers["keyup"]:
        raise KeyError(uuid)
    self._key_event_handlers["keyup"].pop(uuid)
    if not bool(self._key_event_handlers["keyup"]):
        # dict is empty
        set_key_event_handler = server_pb2.SetKeyEventHandler()
        set_key_event_handler.keyup = False
        obj = server_pb2.ServerCommand()
        obj.set_key_event_handler.CopyFrom(set_key_event_handler)
        self._send_data(obj, uuid)
        ret = self._wait_until(uuid)
        if ret.result.HasField("failure"):
            raise RuntimeError(ret.result.failure)


def add_keydown_handler(
    self: PointCloudViewer,
    handler: Callable[[KeyboardEvent, UUID], None]
) -> UUID:
    """ブラウザでkeydownイベントが発生したときに呼ばれるハンドラーを登録する。

    Args:
        handler (Callable[[KeyboardEvent, UUID], None]): ブラウザでkeydownイベントが発生したときに呼ばれるハンドラー

    Returns:
        UUID: ハンドラーに対応するID。後から操作する際に使う
    """
    uuid = uuid4()
    if "keydown" not in self._key_event_handlers:
        self._key_event_handlers["keydown"] = {}
    self._key_event_handlers["keydown"][uuid] = handler

    set_key_event_handler = server_pb2.SetKeyEventHandler()
    set_key_event_handler.keydown = True
    obj = server_pb2.ServerCommand()
    obj.set_key_event_handler.CopyFrom(set_key_event_handler)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    return uuid


def remove_keydown_handler(
    self: PointCloudViewer,
    uuid: UUID
) -> None:
    """ブラウザでkeydownイベントが発生したときに呼ばれるハンドラーを削除する。

    Args:
        uuid (UUID): ブラウザでkeydownイベントが発生したときに呼ばれるハンドラーのID
    """
    if "keydown" not in self._key_event_handlers:
        raise KeyError(uuid)
    if uuid not in self._key_event_handlers["keydown"]:
        raise KeyError(uuid)
    self._key_event_handlers["keydown"].pop(uuid)
    if not bool(self._key_event_handlers["keydown"]):
        # dict is empty
        set_key_event_handler = server_pb2.SetKeyEventHandler()
        set_key_event_handler.keydown = False
        obj = server_pb2.ServerCommand()
        obj.set_key_event_handler.CopyFrom(set_key_event_handler)
        self._send_data(obj, uuid)
        ret = self._wait_until(uuid)
        if ret.result.HasField("failure"):
            raise RuntimeError(ret.result.failure)


def add_keypress_handler(
    self: PointCloudViewer,
    handler: Callable[[KeyboardEvent, UUID], None]
) -> UUID:
    """ブラウザでkeypressイベントが発生したときに呼ばれるハンドラーを登録する。

    Args:
        handler (Callable[[KeyboardEvent, UUID], None]): ブラウザでkeypressイベントが発生したときに呼ばれるハンドラー

    Returns:
        UUID: ハンドラーに対応するID。後から操作する際に使う
    """
    uuid = uuid4()
    if "keypress" not in self._key_event_handlers:
        self._key_event_handlers["keypress"] = {}
    self._key_event_handlers["keypress"][uuid] = handler

    set_key_event_handler = server_pb2.SetKeyEventHandler()
    set_key_event_handler.keypress = True
    obj = server_pb2.ServerCommand()
    obj.set_key_event_handler.CopyFrom(set_key_event_handler)
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    return uuid


def remove_keypress_handler(
    self: PointCloudViewer,
    uuid: UUID
) -> None:
    """ブラウザでkeypressイベントが発生したときに呼ばれるハンドラーを削除する。

    Args:
        uuid (UUID): ブラウザでkeypressイベントが発生したときに呼ばれるハンドラーのID
    """
    if "keypress" not in self._key_event_handlers:
        raise KeyError(uuid)
    if uuid not in self._key_event_handlers["keypress"]:
        raise KeyError(uuid)
    self._key_event_handlers["keypress"].pop(uuid)
    if not bool(self._key_event_handlers["keypress"]):
        # dict is empty
        set_key_event_handler = server_pb2.SetKeyEventHandler()
        set_key_event_handler.keypress = False
        obj = server_pb2.ServerCommand()
        obj.set_key_event_handler.CopyFrom(set_key_event_handler)
        self._send_data(obj, uuid)
        ret = self._wait_until(uuid)
        if ret.result.HasField("failure"):
            raise RuntimeError(ret.result.failure)



import multiprocessing
import pkgutil
import asyncio
import threading

from http.server import BaseHTTPRequestHandler, HTTPServer
from os.path import join
from typing import Union, Optional

import websockets


def _MakePointCloudViewerHTTPRequestHandler(websocket_port: int):
    class _PointCloudViewerHTTPRequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            if self.path == "/":
                self.redirect("index.html")
            elif self.path == "/websocket_url":
                self.send_response(200)
                self.end_headers()
                self.wfile.write(
                    f"ws://{self.server.server_address[0]}:{websocket_port}".encode("utf-8")
                )
            else:
                path = join("/public/", "./"+self.path)
                data: Union[bytes, None] = None
                try:
                    data = pkgutil.get_data("cumo", path)
                except FileNotFoundError:
                    self.send_error(404)
                if data is not None:
                    self.send_response(200)
                    self.end_headers()
                    self.wfile.write(data)
                else:
                    self.send_error(404)

        # pylint: disable=W0622
        def log_message(self, format: str, *args) -> None:
            pass

        def redirect(self, path: str):
            self.send_response(301)
            location = path
            self.send_header(
                "Location", location
            )
            self.end_headers()
    return _PointCloudViewerHTTPRequestHandler


def multiprocessing_worker(
    host: str,
    websocket_port: int,
    http_port: int,
    websocket_broadcasting_queue: multiprocessing.Queue,
    websocket_message_queue: multiprocessing.Queue,
):
    websocket_connection: Optional[websockets.WebSocketServerProtocol] = None

    async def __broadcast():
        nonlocal websocket_connection
        loop = asyncio.get_running_loop()
        try:
            while True:
                data = await loop.run_in_executor(None, websocket_broadcasting_queue.get)
                await websocket_connection.send(data)
        except asyncio.CancelledError:
            pass

    async def __websocket_handler(websocket: websockets.WebSocketServerProtocol, _path: str):
        nonlocal websocket_connection
        if websocket_connection is not None:
            await websocket.close()
            return
        websocket_connection = websocket
        loop = asyncio.get_running_loop()
        broadcast_task = loop.create_task(__broadcast())
        try:
            async for msg in websocket:
                msg: bytes = msg
                websocket_message_queue.put(msg)
        finally:
            websocket_connection = None
            broadcast_task.cancel()

    loop = asyncio.get_event_loop()
    http_server = HTTPServer(
        (host, http_port),
        _MakePointCloudViewerHTTPRequestHandler(websocket_port=websocket_port),
    )

    start_server = websockets.serve(__websocket_handler,
                                    host=host,
                                    port=websocket_port,
                                    max_size=None)
    loop.run_until_complete(start_server)

    threads = [
        threading.Thread(
            target=loop.run_forever
        ),
        threading.Thread(
            target=http_server.serve_forever
        )
    ]
    for thread in threads:
        thread: threading.Thread = thread
        thread.setDaemon(True)
        thread.start()

    for thread in threads:
        thread.join()



class KeyboardEvent:
    key: str
    code: str
    shiftKey: bool
    altKey: bool
    ctrlKey: bool
    metaKey: bool
    repeat: bool

    def __init__(
        self,
        key: str,
        code: str,
        shiftKey: bool,
        altKey: bool,
        ctrlKey: bool,
        metaKey: bool,
        repeat: bool
    ) -> None:
        self.key = key
        self.code = code
        self.shiftKey = shiftKey
        self.altKey = altKey
        self.ctrlKey = ctrlKey
        self.metaKey = metaKey
        self.repeat = repeat

    def __str__(self) -> str:
        return f"""KeyboardEvent(key: \"{
            self.key
        }\", code: \"{
            self.code
        }\", shiftKey: {
            self.shiftKey
        }, altKey: {
            self.altKey
        }, ctrlKey: {
            self.ctrlKey
        }, metaKey: {
            self.metaKey
        }, repeat: {
            self.repeat
        })"""



from __future__ import annotations  # Postponed Evaluation of Annotations
from typing import TYPE_CHECKING, Callable, Optional
from uuid import UUID, uuid4
from cumo._internal.protobuf import server_pb2
if TYPE_CHECKING:
    from cumo import PointCloudViewer


# pylint: disable=no-member,redefined-builtin

def set_custom_slider(
    self: PointCloudViewer,
    target: UUID,
    name: Optional[str] = None,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
    step: Optional[float] = None,
    value: Optional[float] = None,
    on_changed: Optional[Callable[[float], None]] = None,
) -> UUID:
    """指定されたスライダーの値を変更する。

    :param target: スライダーの UUID
    :type target: UUID
    :param name: 表示名
    :type name: Optional[str], optional
    :param min_value: 最小値
    :type min_value: Optional[float], optional
    :param max_value: 最大値
    :type max_value: Optional[float], optional
    :param step: 最小の変化量
    :type step: Optional[float], optional
    :param value: 値
    :type value: Optional[float], optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にスライダーの設定値が渡される
    :type on_changed: Optional[Callable[[float], None]], optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    obj = server_pb2.ServerCommand()
    slider = server_pb2.SetCustomControl.Slider()
    if name is not None:
        slider.name = name
    if min_value is not None:
        slider.min = min_value
    if max_value is not None:
        slider.max = max_value
    if step is not None:
        slider.step = step
    if value is not None:
        slider.value = value
    set_custom_control = server_pb2.SetCustomControl()
    set_custom_control.slider.CopyFrom(slider)
    set_custom_control.target = str(target)
    obj.set_custom_control.CopyFrom(set_custom_control)
    if on_changed is not None:
        self._set_custom_handler(target, "changed", on_changed)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def set_custom_checkbox(
    self: PointCloudViewer,
    target: UUID,
    name: Optional[str] = None,
    value: Optional[bool] = None,
    on_changed: Optional[Callable[[bool], None]] = None,
) -> UUID:
    """指定されたチェックボックスの値を変更する。

    :param target: チェックボックスの UUID
    :type target: UUID
    :param name: 表示名
    :type name: Optional[str], optional
    :param value: 値
    :type value: Optional[bool], optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にチェックボックスの設定値が渡される
    :type on_changed: Optional[Callable[[bool], None]], optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    checkbox = server_pb2.SetCustomControl.CheckBox()
    if name is not None:
        checkbox.name = name
    if value is not None:
        checkbox.value = value
    set_custom_control = server_pb2.SetCustomControl()
    set_custom_control.checkbox.CopyFrom(checkbox)
    set_custom_control.target = str(target)
    obj = server_pb2.ServerCommand()
    obj.set_custom_control.CopyFrom(set_custom_control)
    if on_changed is not None:
        self._set_custom_handler(target, "changed", on_changed)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def set_custom_textbox(
    self: PointCloudViewer,
    target: UUID,
    name: Optional[str] = None,
    value: Optional[str] = None,
    on_changed: Optional[Callable[[bool], None]] = None,
) -> UUID:
    """指定されたテキストボックスの値を変更する。

    :param target: テキストボックスの UUID
    :type target: UUID
    :param name: 表示名
    :type name: Optional[str], optional
    :param value: 値
    :type value: Optional[str], optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にテキストボックスの設定値が渡される
    :type on_changed: Optional[Callable[[bool], None]], optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    textbox = server_pb2.SetCustomControl.TextBox()
    if name is not None:
        textbox.name = name
    if value is not None:
        textbox.value = value
    set_custom_control = server_pb2.SetCustomControl()
    set_custom_control.textbox.CopyFrom(textbox)
    set_custom_control.target = str(target)
    obj = server_pb2.ServerCommand()
    obj.set_custom_control.CopyFrom(set_custom_control)
    if on_changed is not None:
        self._set_custom_handler(target, "changed", on_changed)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def set_custom_selectbox(
    self: PointCloudViewer,
    target: UUID,
    items: Optional[list] = None,
    name: Optional[str] = None,
    value: Optional[int] = None,
    on_changed: Optional[Callable[[int], None]] = None,
) -> UUID:
    """指定されたセレクトボックスの値を変更する。

    :param target: セレクトボックスの UUID
    :type target: UUID
    :param items: 要素の文字列のリスト
    :type items: Optional[list], optional
    :param name: 表示名
    :type name: Optional[str], optional
    :param value: 値
    :type value: Optional[int], optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数。引数にセレクトボックスの設定値が渡される
    :type on_changed: Optional[Callable[[int], None]], optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    selectbox = server_pb2.SetCustomControl.SelectBox()
    if name is not None:
        selectbox.name = name
    if value is not None:
        selectbox.value = value
    if items is not None:
        selectbox.items.extend(items)
    set_custom_control = server_pb2.SetCustomControl()
    set_custom_control.selectbox.CopyFrom(selectbox)
    set_custom_control.target = str(target)
    obj = server_pb2.ServerCommand()
    obj.set_custom_control.CopyFrom(set_custom_control)
    if on_changed is not None:
        self._set_custom_handler(target, "changed", on_changed)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def set_custom_button(
    self: PointCloudViewer,
    target: UUID,
    name: Optional[str] = None,
    on_changed: Optional[Callable[[bool], None]] = None,
) -> UUID:
    """指定されたボタンの値を変更する。

    :param target: ボタンの UUID
    :type target: UUID
    :param name: 表示名
    :type name: Optional[str], optional
    :param on_changed: ボタンがクリックされたときに呼ばれるコールバック関数
    :type on_changed: Optional[Callable[[bool], None]], optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    button = server_pb2.SetCustomControl.Button()
    if name is not None:
        button.name = name
    set_custom_control = server_pb2.SetCustomControl()
    set_custom_control.button.CopyFrom(button)
    set_custom_control.target = str(target)
    obj = server_pb2.ServerCommand()
    obj.set_custom_control.CopyFrom(set_custom_control)
    if on_changed is not None:
        self._set_custom_handler(target, "changed", on_changed)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def set_custom_colorpicker(
    self: PointCloudViewer,
    target: UUID,
    name: Optional[str] = None,
    value: Optional[str] = None,
    on_changed: Optional[Callable[[], None]] = None,
) -> UUID:
    """指定されたフォルダーのカラーピッカーを変更する。

    :param target: フォルダーの UUID
    :type target: UUID
    :param name: 表示名
    :type name: Optional[str], optional
    :param value: 値
    :type value: Optional[str], optional
    :param on_changed: 値が変化したときに呼ばれるコールバック関数
    :type on_changed: Optional[Callable[[], None]], optional

    Returns:
        UUID: コントロールに対応するID。後から操作する際に使う
    """
    picker = server_pb2.SetCustomControl.ColorPicker()
    if name is not None:
        picker.name = name
    if value is not None:
        picker.value = value
    set_custom_control = server_pb2.SetCustomControl()
    set_custom_control.color_picker.CopyFrom(picker)
    set_custom_control.target = str(target)
    obj = server_pb2.ServerCommand()
    obj.set_custom_control.CopyFrom(set_custom_control)
    if on_changed is not None:
        self._set_custom_handler(target, "changed", on_changed)
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)



from __future__ import annotations  # Postponed Evaluation of Annotations
from typing import TYPE_CHECKING
from uuid import UUID, uuid4
from cumo._internal.protobuf import server_pb2
if TYPE_CHECKING:
    from cumo import PointCloudViewer


# pylint: disable=no-member

def remove_all_objects(
    self: PointCloudViewer
) -> None:
    """すべてのオブジェクトとオーバーレイを削除する。
    """
    remove_object_cmd = server_pb2.RemoveObject()
    remove_object_cmd.all = True

    obj = server_pb2.ServerCommand()
    obj.remove_object.CopyFrom(remove_object_cmd)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def remove_object(
    self: PointCloudViewer,
    uuid: UUID
) -> None:
    """指定したUUIDを持つオブジェクトやオーバーレイを削除する。

    Args:
        uuid (UUID): オブジェクトやオーバーレイのUUID
    """
    remove_object_cmd = server_pb2.RemoveObject()
    remove_object_cmd.by_uuid = str(uuid)

    obj = server_pb2.ServerCommand()
    obj.remove_object.CopyFrom(remove_object_cmd)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)



from __future__ import annotations  # Postponed Evaluation of Annotations
from typing import TYPE_CHECKING
from uuid import uuid4
from cumo._internal.protobuf import server_pb2
if TYPE_CHECKING:
    from cumo import PointCloudViewer


def wait_forever(self: PointCloudViewer) -> None:
    """
    サーバーが動作している間待ち続ける。
    この関数を実行した後にコールバック内で ``sys.exit()`` を呼び出すなどすることでプログラムを終了できる。
    """
    self._wait_until(None)


def console_log(
    self: PointCloudViewer,
    message: str,
) -> None:
    """ブラウザ上で ``console.log()`` を実行する。

    :param message: ``console.log()`` に引数として渡される文字列
    :type message: str
    """
    obj = server_pb2.ServerCommand()
    obj.log_message = message
    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)


def start(self: PointCloudViewer) -> None:
    """
    サーバープロセスを起動する。
    """

    self._server_process.start()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cumo',
 'cumo._internal',
 'cumo._internal.members',
 'cumo._internal.protobuf',
 'cumo._vendor.pypcd']

package_data = \
{'': ['*'], 'cumo': ['public/*']}

install_requires = \
['Pillow>=9.0.1,<10.0.0',
 'numpy>=1.20.2,<2.0.0',
 'protobuf>=3.15.6,<4.0.0',
 'python-lzf>=0.2.4,<0.3.0',
 'websockets>=8.1,<9.0']

setup_kwargs = {
    'name': 'cumo',
    'version': '0.21.0',
    'description': 'Webブラウザ上に点群を描画する python ライブラリ',
    'long_description': '# cumo\n\n## ビルド\n\nyarn、protoc、poetryが必要です（`.devcontainer/Dockerfile`参照）。\n以下のようにすると`lib/dist`にtar.gzとwhlファイルが生成されます。\nクライアントのHTML等はライブラリに埋め込まれています。\n\n```console\n$ ./build.sh\n```\n\n## インストール\n\n```console\n$ pip install pointcloud-viewer\n```\n\n## ドキュメント\n\nsphinxでドキュメントの生成が可能です。\n\n```console\n$ cd lib\n$ poetry install\n$ poetry run sphinx-apidoc --append-syspath -F -o ./docs .\n```\n\n## 使用例\n\n`lib/cumo/__main__.py`は3面図を撮る例です。\n\n```console\n$ poetry run python -m cumo pcl_logo.pcd\nopen: http://127.0.0.1:8082\nsetup...\nresize window and press custom control button "start"\nsaved: screenshot_x.png\nsaved: screenshot_y.png\nsaved: screenshot_z.png\n```\n\nREPLでの使用も可能です。\n\n```console\n$ poetry run python\nPython 3.8.7 (default, Apr  9 2022, 21:34:33)\n[GCC 9.4.0] on linux\nType "help", "copyright", "credits" or "license" for more information.\n>>> from cumo import PointCloudViewer\n>>> viewer = PointCloudViewer()\n>>> viewer.start()\n>>> # open localhost:8082 on your browser\n>>> with open(filename, "rb") as f:\n>>>     b = f.read()\n>>>     viewer.send_pointcloud_pcd(b)\n```\n',
    'author': 'Kurusugawa Computer',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kurusugawa-computer/cumo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



from __future__ import annotations  # Postponed Evaluation of Annotations
from typing import TYPE_CHECKING
import multiprocessing
from cumo._internal.server import multiprocessing_worker
if TYPE_CHECKING:
    from cumo import PointCloudViewer


# pylint: disable=no-member

def __init__(
    self: PointCloudViewer,
    host: str = "127.0.0.1",
    websocket_port: int = 8081,
    http_port: int = 8082,
    autostart: bool = False
) -> None:
    self._custom_handlers = {}
    self._key_event_handlers = {}
    self._websocket_broadcasting_queue = multiprocessing.Queue()
    self._websocket_message_queue = multiprocessing.Queue()
    self._server_process = multiprocessing.Process(
        target=multiprocessing_worker,
        args=(
            host,
            websocket_port,
            http_port,
            self._websocket_broadcasting_queue,
            self._websocket_message_queue
        ),
        daemon=True
    )
    if autostart:
        self.start()



import sys
from uuid import UUID
from argparse import ArgumentParser
import numpy
from cumo import PointCloudViewer
from cumo.keyboard_event import KeyboardEvent


def main():
    host = "127.0.0.1"
    websocket_port = 8081
    http_port = 8082

    parser = ArgumentParser()
    parser.add_argument("pcd_filepath")
    args = parser.parse_args()

    viewer = PointCloudViewer(
        host=host, websocket_port=websocket_port, http_port=http_port, autostart=True
    )
    print(f"open: http://{host}:{http_port}")
    print("setup...")

    viewer.remove_all_objects()
    viewer.remove_all_custom_controls()

    radius = send_pointcloud_pcd(viewer, args.pcd_filepath)

    points = numpy.array([
        [0, 0, 0],
        [1, 0, 0], [0, 1, 0], [0, 0, 1],
        [1, 1, 0], [1, 0, 1], [0, 1, 1],
        [2, 0, 0], [0, 2, 0], [0, 0, 2]
    ])
    points = points * numpy.ones((3,))*0.1
    points = points.astype("float32")

    lines = numpy.array([
        [0, 7], [0, 8], [0, 9]
    ]).astype("uint32")
    colors = numpy.array([
        [255, 0, 0], [0, 255, 0], [0, 0, 255],
    ]).astype("uint8")
    widths = (numpy.ones((3,))*5.0).astype("float32")
    viewer.send_lineset(points, lines, colors, widths)

    viewer.send_overlay_text(
        "x", points[7][0], points[7][1], points[7][2], style="color: red"
    )
    viewer.send_overlay_text(
        "y", points[8][0], points[8][1], points[8][2], style="color: green"
    )
    viewer.send_overlay_text(
        "z", points[9][0], points[9][1], points[9][2], style="color: blue"
    )
    viewer.send_overlay_text(args.pcd_filepath, 10, 10,
                             screen_coordinate=True, style="font-family: monospace")

    triangles = numpy.array([
        [0, 1, 2], [0, 2, 3], [3, 1, 0]
    ]).astype("uint32")
    colors = numpy.array([
        [255, 255, 255], [255, 0, 0], [0, 255, 0], [0, 0, 255]
    ]).astype("uint8")
    viewer.send_mesh(points, triangles, colors)

    viewer.set_orthographic_camera(frustum_height=radius*2)

    def save_png(name: str, data: bytes) -> None:
        with open(name, "bw") as f:
            f.write(data)
            f.close()
            print("saved: "+name)

    def take_screenshots():
        viewer.set_camera_position(0, 0, 1)
        overlay_data = viewer.capture_screen()
        viewer.send_overlay_image(overlay_data, 300, radius, 0, 0)

        for p in [(1, 0, 0, "screenshot_x.png"), (0, 1, 0, "screenshot_y.png"), (0, 0, 1, "screenshot_z.png")]:
            viewer.set_camera_position(p[0], p[1], p[2])
            data = viewer.capture_screen()
            save_png(p[3], data)

    def on_keyup(ev: KeyboardEvent, handler_id: UUID):
        if ev.code == "KeyA":
            take_screenshots()
            viewer.remove_keyup_handler(handler_id)
            sys.exit(0)

    keyup_handler_id = viewer.add_keyup_handler(on_keyup)

    def on_start_button_pushed():
        take_screenshots()
        viewer.remove_keyup_handler(keyup_handler_id)
        sys.exit(0)

    viewer.add_custom_button(
        name="start", on_changed=lambda dummy: on_start_button_pushed()
    )

    print("resize window and press custom control button \"start\" (or press A key)")

    viewer.wait_forever()


def send_pointcloud_pcd(viewer: PointCloudViewer, filename: str) -> float:
    with open(filename, "rb") as f:
        b = f.read()
        viewer.send_pointcloud_pcd(b)
    return 1


if __name__ == "__main__":
    main()



from __future__ import annotations  # Postponed Evaluation of Annotations
import io
from typing import TYPE_CHECKING, Optional
from uuid import UUID, uuid4
import html
import numpy
from PIL import Image
from numpy import ndarray
from cumo._vendor.pypcd import pypcd
# from pypcd import pypcd
from cumo.pointcloudviewer import DownSampleStrategy
from cumo._internal.protobuf import server_pb2
from cumo._internal.down_sample import down_sample_pointcloud

if TYPE_CHECKING:
    from cumo import PointCloudViewer

DOWNSAMPLING_DEFAULT_MAX_NUM_POINTS = 1_000_000

# pylint: disable=no-member


def send_pointcloud_pcd(
    self: PointCloudViewer,
    pcd_bytes: bytes,
    down_sample: DownSampleStrategy = DownSampleStrategy.RANDOM_SAMPLE,
    max_num_points: int = DOWNSAMPLING_DEFAULT_MAX_NUM_POINTS,
    point_size: float = 1
) -> UUID:
    """点群をブラウザに送信し、表示させる。

    Args:
        pcd_bytes (bytes): pcd形式のデータ。
        down_sample (DownSampleStrategy, optional): DownSampleStrategy.NONE以外を指定すると一定以上の大きさの点群をダウンサンプルする。
            DownSampleStrategy.NONEを指定すると渡されたデータをそのまま送信する。
        max_num_points (int, optional): ダウンサンプルを行う場合、点数をこの数字以下に削減する。
        point_size (int, optional): 点のサイズ。
    Returns:
        UUID: 表示した点群に対応するID。後から操作する際に使う
    """

    if down_sample == DownSampleStrategy.NONE:
        cloud = server_pb2.AddObject.PointCloud()
        cloud.pcd_data = pcd_bytes
        cloud.point_size = point_size

        add_obj = server_pb2.AddObject()
        add_obj.point_cloud.CopyFrom(cloud)

        obj = server_pb2.ServerCommand()
        obj.add_object.CopyFrom(add_obj)

        uuid = uuid4()
        self._send_data(obj, uuid)
        ret = self._wait_until(uuid)
        if ret.result.HasField("failure"):
            raise RuntimeError(ret.result.failure)
        if not ret.result.HasField("success"):
            raise RuntimeError("unexpected response")
        return UUID(hex=ret.result.success)

    pypcd_pc = pypcd.point_cloud_from_buffer(pcd_bytes)
    pc_data: numpy.ndarray = pypcd_pc.pc_data

    xyz = numpy.stack([pc_data["x"], pc_data["y"], pc_data["z"]], axis=1)

    rgb_u32: numpy.ndarray = pc_data["rgb"]
    rgb_u32.dtype = "uint32"
    r_u8: numpy.ndarray = ((rgb_u32 & 0xff0000) >> 16).astype("uint8")
    g_u8: numpy.ndarray = ((rgb_u32 & 0x00ff00) >> 8).astype("uint8")
    b_u8: numpy.ndarray = (rgb_u32 & 0x0000ff).astype("uint8")

    rgb = numpy.stack([r_u8, g_u8, b_u8], axis=1)

    return self.send_pointcloud(
        xyz=xyz, rgb=rgb, down_sample=down_sample, max_num_points=max_num_points, point_size=point_size)


def send_pointcloud(
    self: PointCloudViewer,
    xyz: Optional[numpy.ndarray] = None,
    rgb: Optional[numpy.ndarray] = None,
    xyzrgb: Optional[numpy.ndarray] = None,
    down_sample: Optional[DownSampleStrategy] = DownSampleStrategy.RANDOM_SAMPLE,
    max_num_points: int = DOWNSAMPLING_DEFAULT_MAX_NUM_POINTS,
    point_size: float = 1
) -> UUID:
    """点群をブラウザに送信し、表示させる。

    Args:
        xyz (Optional[numpy.ndarray], optional): shape が (num_points,3) で dtype が float32 の ndarray 。各行が点のx,y,z座標を表す。
        rgb (Optional[numpy.ndarray], optional): shape が (num_points,3) で dtype が uint8 の ndarray 。各行が点のr,g,bを表す。
        xyzrgb (Optional[numpy.ndarray], optional): shape が (num_points,3) で dtype が float32 の ndarray 。
            各行が点のx,y,z座標とrgbを表す。rgbは24ビットのrgb値を r<<16 + g<<8 + b のように float32 にエンコードしたもの。
        down_sample (DownSampleStrategy, optional): DownSampleStrategy.NONE以外を指定すると一定以上の大きさの点群をダウンサンプルする。
        max_num_points (int, optional): ダウンサンプルを行う場合、点数をこの数字以下に削減する。
        point_size (int, optional): 点のサイズ。

    Returns:
        UUID: 表示した点群に対応するID。後から操作する際に使う
    """
    # 引数チェック
    if xyz is None and xyzrgb is None:
        raise ValueError("xyz or xyzrgb is required")
    if xyz is not None and not(len(xyz.shape) == 2 and xyz.shape[1] == 3 and xyz.dtype == "float32"):
        raise ValueError(
            "xyz must be float32 array of shape (num_points, 3)"
        )
    if rgb is not None:
        if xyz is None:
            raise ValueError("xyz is required with rgb")
        shape_is_valid = len(rgb.shape) == 2 and rgb.shape[1] == 3
        length_is_same = shape_is_valid and rgb.shape[0] == xyz.shape[0]
        type_is_valid = rgb.dtype == "uint8"

        if not (shape_is_valid and length_is_same and type_is_valid):
            raise ValueError(
                "rgb must be uint8 array of shape (num_points, 3)"
            )
    if xyzrgb is not None and not(len(xyzrgb.shape) == 2 and xyzrgb.shape[1] == 4 and xyzrgb.dtype == "float32"):
        raise ValueError(
            "xyzrgb must be float32 array of shape (num_points, 4)"
        )

    # pcdデータ作成
    pcd: pypcd.PointCloud
    if xyz is not None:
        if rgb is not None:
            rgb_u32 = rgb.astype("uint32")
            rgb_f32: numpy.ndarray = (
                (rgb_u32[:, 0] << 16)
                + (rgb_u32[:, 1] << 8)
                + rgb_u32[:, 2]
            )
            rgb_f32.dtype = "float32"

            concatenated: numpy.ndarray = numpy.column_stack((
                xyz,
                rgb_f32,
            ))
            pcd = pypcd.make_xyz_rgb_point_cloud(
                down_sample_pointcloud(
                    concatenated, down_sample, max_num_points=max_num_points)
            )
        else:
            pcd = pypcd.make_xyz_point_cloud(
                down_sample_pointcloud(
                    xyz, down_sample, max_num_points=max_num_points)
            )
    else:
        assert xyzrgb is not None
        pcd = pypcd.make_xyz_rgb_point_cloud(
            down_sample_pointcloud(xyzrgb, down_sample,
                                   max_num_points=max_num_points)
        )

    pcd_bytes = pcd.save_pcd_to_buffer()

    # 送信
    return self.send_pointcloud_pcd(pcd_bytes, down_sample=DownSampleStrategy.NONE, point_size=point_size)


# pylint: disable=too-many-branches
def send_lineset(
    self: PointCloudViewer,
    xyz: numpy.ndarray,
    from_to: numpy.ndarray,
    rgb: Optional[numpy.ndarray] = None,
    width: Optional[numpy.ndarray] = None
) -> UUID:
    """Linesetをブラウザに送信し、表示させる。

    Args:
        xyz (numpy.ndarray): shape が (num_points,3) で dtype が float32 の ndarray 。各行が線分の端点のx,y,z座標を表す。
        from_to (numpy.ndarray): shape が (num_lines,2) で dtype が uint32 の ndarray 。各行が線分の端点のインデックスによって1本の線分を表す。
        rgb (Optional[numpy.ndarray], optional): shape が (num_lines,3) で dtype が uint8 の ndarray 。各行が線分のr,g,bを表す。
        width (Optional[numpy.ndarray], optional): shape が (num_lines,) で dtypeが float32 の ndarray 。各要素が線分の太さを表す。
    Returns:
        UUID: 表示したLinesetに対応するID。後から操作する際に使う
    """
    if not (len(xyz.shape) == 2 and xyz.shape[1] == 3 and xyz.dtype == "float32"):
        raise ValueError("xyz must be float32 array of shape (num_points,3)")
    if not (len(from_to.shape) == 2 and from_to.shape[1] == 2 and from_to.dtype == "uint32"):
        raise ValueError("from_to must be uint32 array of shape (num_lines,2)")

    if rgb is not None:
        shape_is_valid = len(rgb.shape) == 2 and rgb.shape[1] == 3
        type_is_valid = rgb.dtype == "uint8"

        if not (shape_is_valid and type_is_valid):
            raise ValueError(
                "rgb must be uint8 array of shape (num_lines, 3)"
            )
    if width is not None:
        shape_is_valid = len(width.shape) == 1
        if rgb is not None:
            shape_is_valid = shape_is_valid and rgb.shape[0] == width.shape[0]
        type_is_valid = width.dtype == "float32"

        if not (shape_is_valid and type_is_valid):
            raise ValueError(
                "whidth must be float32 array of shape (num_lines,)"
            )

    num_points = xyz.shape[0]

    pb_lineset = server_pb2.AddObject.LineSet()
    for v in xyz:
        p = server_pb2.VecXYZf()
        p.x = v[0]
        p.y = v[1]
        p.z = v[2]
        pb_lineset.points.append(p)
    for l in from_to:
        if not (0 <= l[0] and l[0] < num_points and 0 <= l[1] and l[1] < num_points):
            raise ValueError(
                "value of from_to element must be 0 <= and < num_points")
        pb_lineset.from_index.append(l[0])
        pb_lineset.to_index.append(l[1])

    if rgb is not None:
        for l in rgb:
            c = server_pb2.VecRGBf()
            c.r = l[0]
            c.g = l[1]
            c.b = l[2]
            pb_lineset.colors.append(c)
    if width is not None:
        for w in width:
            pb_lineset.widths.append(w)
    add_obj = server_pb2.AddObject()
    add_obj.line_set.CopyFrom(pb_lineset)
    obj = server_pb2.ServerCommand()
    obj.add_object.CopyFrom(add_obj)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def send_mesh(
    self: PointCloudViewer,
    xyz: numpy.ndarray,
    indices: numpy.ndarray,
    rgb: Optional[numpy.ndarray] = None
) -> UUID:
    """Meshをブラウザに送信し、表示させる。

    Args:
        xyz (numpy.ndarray): shape が (num_points,3) で dtype が float32 の ndarray 。各行が頂点のx,y,z座標を表す。
        indices (numpy.ndarray): shape が (num_triangles,3) で dtype が uint32 の ndarray 。各行が頂点のインデックスによって1枚の三角形を表す。
        rgb (Optional[numpy.ndarray], optional): shape が (num_points,3) で dtype が uint8 の ndarray 。各行が頂点のr,g,bを表す。

    Returns:
        UUID: 表示したMeshに対応するID。後から操作する際に使う
    """
    if not (len(xyz.shape) == 2 and xyz.shape[1] == 3 and xyz.dtype == "float32"):
        raise ValueError("xyz must be float32 array of shape (num_points,3)")
    if not (len(indices.shape) == 2 and indices.shape[1] == 3 and indices.dtype == "uint32"):
        raise ValueError(
            "indices must be uint32 array of shape (num_triangles,3)"
        )
    if rgb is not None:
        shape_is_valid = len(rgb.shape) == 2 and rgb.shape[1] == 3
        type_is_valid = rgb.dtype == "uint8"

        if not (shape_is_valid and type_is_valid):
            raise ValueError(
                "rgb must be uint8 array of shape (num_triangles, 3)"
            )

    num_points = xyz.shape[0]

    pb_mesh = server_pb2.AddObject.Mesh()
    for v in xyz:
        p = server_pb2.VecXYZf()
        p.x = v[0]
        p.y = v[1]
        p.z = v[2]
        pb_mesh.points.append(p)
    for l in indices:
        for i in range(3):
            if not (0 <= l[i] and l[i] < num_points):
                raise ValueError(
                    "value of indices element must be 0 <= and < num_points")
        pb_mesh.vertex_a_index.append(l[0])
        pb_mesh.vertex_b_index.append(l[1])
        pb_mesh.vertex_c_index.append(l[2])
    if rgb is not None:
        rgb_f = rgb.astype("float32")
        rgb_f /= 255
        for l in rgb_f:
            c = server_pb2.VecRGBf()
            c.r = l[0]
            c.g = l[1]
            c.b = l[2]
            pb_mesh.colors.append(c)
    add_obj = server_pb2.AddObject()
    add_obj.mesh.CopyFrom(pb_mesh)
    obj = server_pb2.ServerCommand()
    obj.add_object.CopyFrom(add_obj)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def send_overlay_text(
    self: PointCloudViewer,
    text: str,
    x: float = 0,
    y: float = 0,
    z: float = 0,
    screen_coordinate: bool = False,
    style: str = "",
) -> UUID:
    """特定の座標を左上として文字列をオーバーレイさせる。

    :param text: 表示させる文字列
    :type text: str
    :param x: オーバーレイが追従する点のx座標
    :type x: float, optional
    :param y: オーバーレイが追従する点のy座標
    :type y: float, optional
    :param z: オーバーレイが追従する点のz座標
    :type z: float, optional
    :param screen_coordinate: Trueにするとオーバーレイが画面の指定の位置に固定される。このときzは無視される
    :type screen_coordinate: bool, optional
    :param style: style属性に渡される文字列
    :type style: str, optional

    Returns:
        UUID: オーバーレイに対応するID。後から操作する際に使う
    """
    overlay = server_pb2.AddObject.Overlay()
    position = server_pb2.VecXYZf()
    position.x = x
    position.y = y
    position.z = z
    overlay.position.CopyFrom(position)

    inner_html = html.escape(text).replace("\n", "<br />\n")
    attributes = f"style=\"color:white;mix-blend-mode:difference;{style}\""
    overlay.html = f"<div {attributes}>{inner_html}</div>"

    if screen_coordinate:
        overlay.type = server_pb2.AddObject.Overlay.CoordinateType.SCREEN_COORDINATE
    else:
        overlay.type = server_pb2.AddObject.Overlay.CoordinateType.WORLD_COORDINATE
    add_obj = server_pb2.AddObject()
    add_obj.overlay.CopyFrom(overlay)
    obj = server_pb2.ServerCommand()
    obj.add_object.CopyFrom(add_obj)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)


def send_overlay_image_from_ndarray(
    self: PointCloudViewer,
    ndarray_data: ndarray,
    width: int,
    x: float = 0,
    y: float = 0,
    z: float = 0,
    screen_coordinate: bool = False,
) -> UUID:
    """特定の座標を左上として画像をオーバーレイさせる。

    Args:
        ndarray_data (ndarray): shape が (height,width,3) で dtype が uint8 の ndarray。非圧縮の画像データ。
        width (int): オーバーレイの幅をピクセルで指定
        x (float, optional): オーバーレイが追従する点のx座標
        y (float, optional): オーバーレイが追従する点のy座標
        z (float, optional): オーバーレイが追従する点のz座標
        screen_coordinate (bool, optional) Trueにするとオーバーレイが画面の指定の位置に固定される。このときzは無視される

    Returns:
        UUID: オーバーレイに対応するID。後から操作する際に使う
    """
    if not(len(ndarray_data.shape) == 3 and ndarray_data.shape[2] == 3 and ndarray_data.dtype == "uint8"):
        raise ValueError("ndarray_data must be uint8 array of shape (height, width, 3)")
    img = Image.fromarray(ndarray_data)
    img_bytes = io.BytesIO()
    img.save(img_bytes, format="PNG")
    return self.send_overlay_image(
        img_bytes.getvalue(),
        width,
        x,
        y,
        z,
        screen_coordinate
    )


def send_overlay_image(
    self: PointCloudViewer,
    data: bytes,
    width: int,
    x: float = 0,
    y: float = 0,
    z: float = 0,
    screen_coordinate: bool = False,
) -> UUID:
    """特定の座標を左上として画像をオーバーレイさせる。

    Args:
        data (bytes): 画像データ。jpg,pngに対応
        width (int): オーバーレイの幅をピクセルで指定
        x (float, optional): オーバーレイが追従する点のx座標
        y (float, optional): オーバーレイが追従する点のy座標
        z (float, optional): オーバーレイが追従する点のz座標
        screen_coordinate (bool, optional) Trueにするとオーバーレイが画面の指定の位置に固定される。このときzは無視される

    Returns:
        UUID: オーバーレイに対応するID。後から操作する際に使う
    """
    overlay = server_pb2.AddObject.Overlay()

    position = server_pb2.VecXYZf()
    position.x = x
    position.y = y
    position.z = z
    overlay.position.CopyFrom(position)

    image = server_pb2.AddObject.Overlay.Image()
    image.data = data
    image.width = width
    overlay.image.CopyFrom(image)

    if screen_coordinate:
        overlay.type = server_pb2.AddObject.Overlay.CoordinateType.SCREEN_COORDINATE
    else:
        overlay.type = server_pb2.AddObject.Overlay.CoordinateType.WORLD_COORDINATE

    add_obj = server_pb2.AddObject()
    add_obj.overlay.CopyFrom(overlay)
    obj = server_pb2.ServerCommand()
    obj.add_object.CopyFrom(add_obj)

    uuid = uuid4()
    self._send_data(obj, uuid)
    ret = self._wait_until(uuid)
    if ret.result.HasField("failure"):
        raise RuntimeError(ret.result.failure)
    if not ret.result.HasField("success"):
        raise RuntimeError("unexpected response")
    return UUID(hex=ret.result.success)
